import typing

import hermitage
import sqlalchemy.ext.asyncio

AsyncEngineContract = sqlalchemy.ext.asyncio.AsyncEngine
AsyncConnectionContract = sqlalchemy.ext.asyncio.AsyncConnection


class ReadAsyncConnectionContract(AsyncConnectionContract):
    pass


class WriteAsyncConnectionContract(AsyncConnectionContract):
    pass


class ShutdownContext(typing.TypedDict):
    exc_type: type[Exception]
    exc_val: Exception


class ReadClientContract(typing.Protocol):
    async def __call__(self, invoice: hermitage.notation.Invoice) -> hermitage.notation.View: ...


class WriteClientContract(typing.Protocol):
    async def __call__(self, invoice: hermitage.notation.Invoice) -> hermitage.notation.View | None: ...



import typing
import asyncio

from hermitage.notation.default import (
    Invoice,
    Bucket,
    Upsert,
    MetaElement
)
from ..configuration import Schema
from ..assembling import QueryBuilder
from ..definition import contracts


class WriteExecutor:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.WriteAsyncConnectionContract,
        builder: QueryBuilder
    ):
        self._schema = schema
        self._connection = connection
        self._builder = builder

    async def __call__(self, bucket: Bucket):
        await self._apply_upsert(bucket)
        await self._process_bucket(bucket)

    async def _process_bucket(self, bucket: Bucket):
        await self._connection.execute(self._builder(bucket))

    async def _apply_upsert(self, bucket: Bucket):
        if upsert := typing.cast(self._search_meta(bucket, Upsert), Upsert):
            query = self._builder(
                Bucket(
                    bucket.name,
                    upsert.clause
                )
            )
            await self._connection.execute(query)

    @staticmethod
    def _search_meta(haystack: Bucket, needle: type[MetaElement]) -> MetaElement | None:
        for element in haystack:
            if isinstance(element, needle):
                return element


class WriteClient:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.WriteAsyncConnectionContract,
    ):
        self._executor = WriteExecutor(
            schema,
            connection,
            QueryBuilder(schema)
        )

    async def __call__(
        self,
        invoice: Invoice
    ) -> typing.NoReturn:
        tasks = []
        async with asyncio.TaskGroup() as tg:
            for bucket in invoice:
                tasks.append(tg.create_task(self._executor(bucket)))



import abc
import collections.abc
import operator
import typing

import sqlalchemy
import zodchy
from collections import deque

from hermitage.notation.default import (
    Bucket,
    Item,
    Total,
    Clause,
    Slice,
    ClauseExpression,
    AND,
    OR
)
from .configuration import (
    Schema,
    M2O,
    Address,
    Space,
    TOTAL_QUERY_FIELD
)


class ClauseCompiler(abc.ABC):
    def __init__(
        self,
        schema: Schema,
        space: Space | None = None
    ):
        self._schema = schema
        self._space = space

    def get_instance(self, space: Space) -> typing.Self:
        return type(self)(schema=self._schema, space=space)

    def __call__(self, expression: ClauseExpression) -> sqlalchemy.ClauseElement | None:
        if expression:
            result = self._compile(expression)
            return result

    @abc.abstractmethod
    def _compile(self, expression: ClauseExpression) -> sqlalchemy.ClauseElement: ...


class FilterCompiler(ClauseCompiler):
    @property
    def _operations(self):
        return {
            zodchy.codex.query.EQ: self._simple_clause(operator.eq),
            zodchy.codex.query.NE: self._simple_clause(operator.ne),
            zodchy.codex.query.LE: self._simple_clause(operator.le),
            zodchy.codex.query.LT: self._simple_clause(operator.lt),
            zodchy.codex.query.GE: self._simple_clause(operator.ge),
            zodchy.codex.query.GT: self._simple_clause(operator.gt),
            zodchy.codex.query.IS: lambda v: self._get_column(v).is_(v.operation.value),
            zodchy.codex.query.LIKE: self._like_clause,
            zodchy.codex.query.NOT: self._not_clause,
            zodchy.codex.query.SET: self._set_clause,
            zodchy.codex.query.RANGE: self._range_clause,
        }

    def _compile(self, expression: ClauseExpression) -> sqlalchemy.ClauseElement:
        stack = deque()
        for element in expression:
            if element is AND:
                a = stack.pop()
                b = stack.pop()
                stack.append(sqlalchemy.and_(a, b))
            elif element is OR:
                stack.append(sqlalchemy.or_(stack.pop(), stack.pop()))
            else:
                stack.append(self._operations[type(element.operation)](element))

        return stack.pop()

    def _get_column(self, clause: Clause) -> sqlalchemy.Column:
        return self._schema.get_column(Address(clause.name, self._space))

    def _not_clause(self, clause: Clause):
        if isinstance(clause.operation, zodchy.codex.query.IS):
            return self._get_column(clause).isnot(clause.operation.value)
        elif isinstance(clause.operation, zodchy.codex.query.EQ):
            return operator.ne(self._get_column(clause), clause.operation.value)
        elif isinstance(clause.operation, zodchy.codex.query.LIKE):
            return self._like_clause(
                Clause(clause.name, clause.operation),
                inversion=True
            )
        elif isinstance(clause.operation, zodchy.codex.query.SET):
            return self._set_clause(
                Clause(clause.name, clause.operation),
                inversion=True
            )
        else:
            return sqlalchemy.not_(self._compile(clause)),

    def _simple_clause(self, op):
        return lambda v: op(self._get_column(v), v.operation.value)

    def _logic_clause(self, op):
        return lambda v: op(self._compile(u) for u in v)

    def _like_clause(self, clause: Clause, inversion: bool = False):
        column = self._get_column(clause)
        operation = clause.operation
        value = f'%{operation.value}%'
        if operation.case_sensitive:
            return column.notlike(value) if inversion else column.like(value)
        else:
            return column.notilike(value) if inversion else column.ilike(value)

    def _set_clause(self, clause: Clause, inversion: bool = False):
        value = list(clause.operation.value)
        if inversion:
            return self._get_column(clause).notin_(value)
        else:
            return self._get_column(clause).in_(value)

    def _range_clause(self, clause: Clause):
        params = [
            Clause(clause.name, condition)
            for condition in clause.operation.value
            if condition is not None
        ]
        if len(params) > 1:
            return self._logic_clause(sqlalchemy.and_)(params)
        elif len(params) == 1:
            return params[0]


class OrderCompiler(ClauseCompiler):
    def __call__(self, clause: ClauseExpression) -> list[sqlalchemy.ClauseElement]:
        result = self._compile(clause)
        return result

    @property
    def _operations(self):
        return {
            zodchy.codex.query.ASC: sqlalchemy.asc,
            zodchy.codex.query.DESC: sqlalchemy.desc
        }

    def _compile(self, expression: ClauseExpression) -> list[sqlalchemy.ClauseElement]:
        stack = deque()
        result = []
        for element in expression or ():
            if element is AND:
                while stack:
                    result.append(stack.pop())
            else:
                result.append(
                    self._operations[type(element.operation)](
                        self._schema.get_column(Address(element.name, self._space))
                    )
                )
        return result


class Query:
    def __init__(
        self,
        schema: Schema
    ):
        self._schema = schema
        self._filter_compiler = FilterCompiler(self._schema)
        self._order_compiler = OrderCompiler(self._schema)
        self._aliases = {}
        self._select = []
        self._joins = []
        self._values = []
        self._filters = []
        self._orders = []
        self._limit = None
        self._offset = None

    def __call__(self, bucket: Bucket):
        self._process_bucket(bucket)
        return self._build_query(bucket)

    def _build_query(self, bucket: Bucket):
        if self._select:
            q = sqlalchemy.select(*self._select)
        elif self._values:
            if self._filters:
                q = sqlalchemy.update(self._schema.get_table(Space(bucket.name))).values(self._values[0])
            else:
                q = sqlalchemy.insert(self._schema.get_table(Space(bucket.name))).values(self._values)
        else:
            if self._filters:
                q = sqlalchemy.delete(self._schema.get_table(Space(bucket.name)))
            else:
                raise ValueError("Cannot determine operation type")

        for join in self._joins:
            q = q.join_from(*join, isouter=True)
        for filter_clause in self._filters:
            q = q.where(filter_clause)
        for order_clause in self._orders:
            q = q.order_by(order_clause)
        if self._limit:
            q = q.limit(self._limit)
        if self._offset:
            q = q.offset(self._offset)

        return q

    def _process_bucket(self, bucket: Bucket, parent: Space | None = None):
        bucket_space = Space(bucket.qualified_name)
        if parent:
            bucket_space = parent + bucket_space

        _filter_clause = None
        _order_clause = None
        _limit_clause = None
        _offset_clause = None
        for element in bucket:
            if isinstance(element, Bucket):
                element_space = bucket_space + Space(element.qualified_name)
                if isinstance(link := self._schema.get_link(bucket_space, element_space), M2O):
                    source_column = self._schema.get_column(Address(link.source_address.name, bucket_space))
                    target_column = self._schema.get_column(Address(link.target_address.name, element_space))
                    self._joins.append((
                        self._schema.get_table(bucket_space),
                        self._schema.get_table(element_space),
                        source_column == target_column
                    ))
                    self._process_bucket(element, bucket_space)
            elif isinstance(element, Item):
                _value = element()
                if isinstance(_value, str):
                    self._select.append(
                        self._schema.get_column(Address(_value, bucket_space))
                    )
                elif isinstance(_value, collections.abc.Mapping):
                    self._values.append(_value)
            elif isinstance(element, Total):
                self._select.append(
                    sqlalchemy.text(f"count(*) over () as {TOTAL_QUERY_FIELD}")
                )
            elif isinstance(element, Clause):
                if isinstance(element.operation, zodchy.codex.query.FilterBit):
                    _filter_clause = ClauseExpression(element) if _filter_clause is None else _filter_clause & element
                elif isinstance(element.operation, zodchy.codex.query.OrderBit):
                    _order_clause = ClauseExpression(element) if _order_clause is None else _order_clause & element
            elif isinstance(element, Slice):
                if isinstance(element.operation, zodchy.codex.query.Limit):
                    _limit_clause = element.operation.value
                elif isinstance(element.operation, zodchy.codex.query.Offset):
                    _offset_clause = element.operation.value
            elif isinstance(element, ClauseExpression):
                if isinstance(element[0].operation, zodchy.codex.query.FilterBit):
                    _filter_clause = element if _filter_clause is None else _filter_clause & element
                elif isinstance(element[0].operation, zodchy.codex.query.OrderBit):
                    _order_clause = element if _order_clause is None else _order_clause & element

        if _filter_clause:
            filter_compiler = self._filter_compiler.get_instance(bucket_space)
            self._filters.append(filter_compiler(_filter_clause))

        if _order_clause:
            order_compiler = self._order_compiler.get_instance(bucket_space)
            self._orders.extend(order_compiler(_order_clause))

        if _limit_clause:
            self._limit = _limit_clause

        if _offset_clause:
            self._offset = _offset_clause


class QueryBuilder:
    def __init__(self, schema: Schema):
        self._schema = schema

    def __call__(self, bucket: Bucket):
        return Query(schema=self._schema)(bucket)



import collections.abc
import typing
import asyncio

import zodchy

from hermitage.notation import (
    Invoice,
    Bucket,
    Item,
    Clause,
    View
)
from ..configuration import Schema, O2M, M2M, Space, TOTAL_QUERY_FIELD
from ..assembling import QueryBuilder
from ..definition import contracts


class Squeezer:
    def __init__(
        self,
        separator: str = '__',
        is_collapse_none: bool = True
    ):
        super().__init__()
        self._separator = separator
        self._is_collapse_none = is_collapse_none

    def __call__(
        self,
        data: collections.abc.MutableMapping
    ):
        data = self._nestify(data)
        return self._collapse_none(data)

    def _split_rec(self, k, v, out):
        k, *rest = k.split(self._separator, 1)
        if rest:
            self._split_rec(rest[0], v, out.setdefault(k, {}))
        else:
            out[k] = v

    def _nestify(
        self,
        data: typing.Any
    ):
        if isinstance(data, str):
            return data
        elif isinstance(data, bytearray):
            return data
        elif isinstance(data, bytes):
            return data
        elif isinstance(data, collections.abc.Mapping):
            result = {}
            for k, v in data.items():
                self._split_rec(k, self._nestify(v), result)
            return result
        elif isinstance(data, collections.abc.Iterable):
            return [self._nestify(item) for item in data]
        else:
            return data

    def _collapse_none(self, data: collections.abc.MutableMapping):
        if self._is_collapse_none is False:
            return data
        result = {}
        for k, v in data.items():
            if isinstance(v, collections.abc.MutableMapping):
                v = self._collapse_none(v) if any(_ is not None for _ in v.values()) else None
            if isinstance(v, list):
                v = [self._collapse_none(_) for _ in v]
            result[k] = v
        return result


class ReadExecutor:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.ReadAsyncConnectionContract,
        builder: QueryBuilder,
    ):
        self._schema = schema
        self._connection = connection
        self._builder = builder
        self._squeezer = Squeezer()

    async def __call__(
        self,
        bucket: Bucket,
        is_nested: bool = False
    ) -> View:
        nested_buckets = []
        # Detection of nested buckets and enrichment parent query with data for further nested calculations
        for element in bucket:
            if isinstance(element, Bucket):
                link = self._schema.get_link(
                    Space(f'{bucket.name}:{element.get_qua()}' if element.get_qua() else bucket.name),
                    Space(element.name)
                )
                if isinstance(link, O2M) or isinstance(link, M2M):
                    # Adding to parent bucket link key field
                    bucket += Item(link.source_address.name)
                    nested_buckets.append((element, link, element.substitution_name))

        # Build and run parent bucket
        query = self._builder(bucket)
        result = await self._connection.execute(query)
        bucket_data = result.mappings().fetchall()

        nested_indexed_data = {}
        # Nested buckets calculation
        for nested_bucket, link, substitution_name in nested_buckets:
            if isinstance(link, O2M):
                index_key = (substitution_name, link.source_address.name)
                nested_bucket += Item(link.target_address.name)
                nested_bucket += Clause(
                    link.target_address.name,
                    zodchy.codex.query.SET(*(e[link.source_address.name] for e in bucket_data))
                )
                nested_view = await self(nested_bucket, True)
                nested_indexed_data[index_key] = self._index_data(
                    data=nested_view.data,
                    key=link.target_address.name,
                    transformer=lambda r: {
                        k: v
                        for k, v in r.items()
                        if k != link.target_address.name
                    }
                )
            elif isinstance(link, M2M):
                index_key = (substitution_name, link.source_address.name)
                nested_bucket = Bucket(
                    link.interim_source_address.space.last,
                    Item(link.interim_source_address.name),
                    Clause(
                        link.interim_source_address.name,
                        zodchy.codex.query.SET(*(e[link.source_address.name] for e in bucket_data))
                    ),
                    nested_bucket + Item(link.target_address.name)
                )
                nested_view = await self(nested_bucket, True)
                nested_indexed_data[index_key] = self._index_data(
                    data=nested_view.data,
                    key=link.interim_source_address.name,
                    transformer=lambda r: {
                        k.replace(f'{substitution_name}__', ''): v
                        for k, v in r.items()
                        if k != link.interim_source_address.name
                    }
                )

        data = []
        meta = None
        for row in bucket_data:
            if is_nested and nested_indexed_data:
                row = dict(row)
            else:
                row = self._squeezer(row)
                if TOTAL_QUERY_FIELD in row:
                    if meta is None:
                        meta = {'total': row[TOTAL_QUERY_FIELD]}
                    del row[TOTAL_QUERY_FIELD]
            for k, v in nested_indexed_data.items():
                row[k[0]] = v.get(row.get(k[1]), [])
            data.append(row)

        return View(data=data, meta=meta)

    @staticmethod
    def _index_data(
        data: collections.abc.Iterable[collections.abc.Mapping],
        key: str,
        transformer: typing.Callable[[collections.abc.Mapping], collections.abc.Mapping] | None = None
    ):
        result = collections.defaultdict(list)
        for row in data:
            index = row[key]
            if transformer:
                row = transformer(row)
            result[index].append(row)
        return result


class ReadClient:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.ReadAsyncConnectionContract
    ):
        self._executor = ReadExecutor(
            schema,
            connection,
            QueryBuilder(schema)
        )

    async def __call__(
        self,
        invoice: Invoice,
    ) -> View | list[View]:
        tasks = []
        async with asyncio.TaskGroup() as tg:
            for bucket in invoice:
                tasks.append(tg.create_task(self._executor(bucket)))
        result = [_.result() for _ in tasks]
        return result if len(result) > 1 else result[0]



import collections.abc
import dataclasses
import typing

import sqlalchemy

TOTAL_QUERY_FIELD = "_total"


@dataclasses.dataclass
class TrackUnit:
    name: str
    qua: str | None = None
    label: str | None = None

    @property
    def qualified_name(self):
        parts = [self.name, self.qua or '', self.label or '']
        return ":".join(parts)

    def __str__(self):
        return self.label or self.qua or self.name


class Space:
    def __init__(self, name: str):
        self._track = []
        for part in name.split('__'):
            if ':' in part:
                a = part.split(':')
                if len(a) > 2:
                    name, qua, label = a
                else:
                    name, qua = a
                    label = None
            else:
                name, qua, label = part, None, None
            self._track.append(TrackUnit(name, qua if qua else None, label if label else None))

    @property
    def qualified_name(self) -> str:
        return "__".join(map(lambda u: u.qualified_name, self._track))

    @property
    def last(self) -> str:
        return self._track[-1].name

    def slice(self, start: int | None, end: int | None) -> typing.Self | None:
        _slice = self._track[start:end]
        if _slice:
            return Space('__'.join(list(map(str, _slice))))

    def __len__(self):
        return len(self._track)

    def __iter__(self):
        yield from self._track

    def __add__(self, other: str | typing.Self):
        if isinstance(other, str):
            return Space(f'{self.qualified_name}__{other}')
        else:
            return Space(f'{self.qualified_name}__{other.qualified_name}')

    def __getitem__(self, item: int):
        return self._track[item]

    def __bool__(self):
        return bool(self._track)

    def __str__(self):
        return '__'.join(list(map(str, self._track)))


class Address:
    def __init__(self, name: str, space: Space):
        self._name = name
        self._space = space

    @property
    def qualified_name(self) -> str:
        return f'{self._space.qualified_name}__{self._name}'

    @property
    def name(self):
        return self._name

    @property
    def space(self):
        return self._space

    def __str__(self):
        return f'{str(self._space)}__{self._name}'


@dataclasses.dataclass
class Link:
    source_address: Address
    target_address: Address


@dataclasses.dataclass
class M2O(Link):
    pass


@dataclasses.dataclass
class O2M(Link):
    pass


@dataclasses.dataclass
class M2M(Link):
    interim_source_address: Address
    interim_target_address: Address


@dataclasses.dataclass(frozen=True)
class Fk:
    target_address: Address
    parent_address: Address

    def __eq__(self, other: TrackUnit) -> bool:
        return other.name == self.target_address.space.last and (
            other.qua is None or f'{other.qua}_id' == self.parent_address.name
        )


class Schema:
    def __init__(
        self,
        tables: collections.abc.Iterable[sqlalchemy.Table],
        *foreign_keys: tuple[sqlalchemy.Column, sqlalchemy.Column]
    ):
        self._graph = collections.defaultdict(set)
        self._tables = {}
        self._cache = {}
        self._index_tables(tables)
        self._index_custom_foreign_keys(*foreign_keys)

    def get_link(
        self,
        source_space: Space,
        target_space: Space
    ):
        link = self._get_m2o(source_space, target_space)
        if not link:
            link = self._get_o2m(source_space, target_space)
        if not link:
            link = self._get_m2m(source_space, target_space)
        return link

    def get_table(self, space: Space) -> sqlalchemy.Table | None:
        if str(space) in self._cache:
            return self._cache[str(space)]
        last_unit = space[-1]
        result = self._tables.get(last_unit.name)
        if last_unit.label or last_unit.qua:
            result = result.alias(str(space))
            self._cache[str(space)] = result
        return result

    def get_column(self, address: Address) -> sqlalchemy.Column | None:
        try:
            address_string = str(address)
            if address_string not in self._cache:
                column = getattr(
                    self.get_table(address.space).c, address.name
                )
                space = address.space.slice(1, None)
                if space:
                    column = column.label(str(Address(address.name, space)))
                self._cache[address_string] = column
            return self._cache[address_string]
        except AttributeError:
            return None

    def _index_tables(
        self,
        tables: collections.abc.Iterable[sqlalchemy.Table]
    ):
        for table in tables:
            self._tables[table.name] = table
            for fk in table.foreign_keys:
                self._graph[fk.parent.table.name].add(
                    Fk(
                        target_address=Address(name=fk.column.name, space=Space(fk.column.table.name)),
                        parent_address=Address(name=fk.parent.name, space=Space(fk.parent.table.name))
                    )
                )

    def _index_custom_foreign_keys(self, *foreign_keys: tuple[sqlalchemy.Column, sqlalchemy.Column]):
        for parent, target in foreign_keys:
            self._graph[parent.table.name].add(
                Fk(
                    parent_address=Address(name=parent.name, space=Space(parent.table.name)),
                    target_address=Address(name=target.name, space=Space(target.table.name))
                )
            )

    def _get_m2o(
        self,
        source_space: Space,
        target_space: Space
    ) -> M2O | None:
        for t, fks in self._graph.items():
            if source_space.last == t:
                for fk in fks:
                    if target_space[-1] == fk:
                        return M2O(
                            source_address=fk.parent_address,
                            target_address=fk.target_address
                        )

    def _get_o2m(
        self,
        source_space: Space,
        target_space: Space,
    ) -> O2M | None:
        for t, fks in self._graph.items():
            if target_space.last == t:
                for fk in fks:
                    if source_space[-1] == fk:
                        return O2M(
                            source_address=fk.target_address,
                            target_address=fk.parent_address
                        )

    def _get_m2m(
        self,
        source_space: Space,
        target_space: Space,
    ) -> M2M | None:
        for t, fks in self._graph.items():
            source_fk = None
            target_fk = None
            for fk in fks:
                if source_space[-1] == fk:
                    source_fk = fk
                if target_space[-1] == fk:
                    target_fk = fk
                if source_fk and target_fk:
                    return M2M(
                        source_address=Address(source_fk.target_address.name, source_space),
                        interim_source_address=Address(source_fk.parent_address.name, Space(t)),
                        target_address=Address(target_fk.target_address.name, target_space),
                        interim_target_address=Address(target_fk.parent_address.name, Space(t)),
                    )
